;Mass storage FATxx DOS-like interface routines

;File handle format:
;Offset:
;	0-3	32-bit size of file
fHandFileSize               equ 0
;	4-5	16-bit starting cluster of file
fHandStartingCluster        equ 4
;	6-7	16-bit current cluster offset for seek pointer
fHandCurClusterOffset       equ 6
;	8	8-bit  current sector offset for seek pointer (first sector of cluster 7-8 would be 0)
fHandCurSectorOffset        equ 8
;	9-12	32-bit seek offset (bytes from start of file)
fHandSeekOffset             equ 9
;	13-14	16-bit current byte offset from start of seek sector
fHandByteOffset             equ 13
;	15-18	32-bit seek sector (adjusted for everything, including partition offset)
fHandSeekAddress            equ 15
;	19-20	16-bit current cluster of the seek pointer
fHandCurCluster             equ 19
;	21	8-bit flag byte
fHandFlags                  equ 21
;	22-25  32-bit LBA address of directory entry
fHandDirEntryAddress        equ 22
;	26     8-bit offset (multiple of 32 bytes) into LBA sector for directory entry
fHandDirEntryOffset         equ 26

 include "settings.inc"
  include "../ti83plus.inc"
 include "equates.inc"
 SEGMENT Main
 GLOBALS ON

 EXTERN getSectorBuffer,FAT_getTimeWord,FAT_getDateWord,UFI_Read,UFI_Write,getMSDPtr,FAT_lookupPath,FAT_nameConvertTo11
 EXTERN FAT_getNextCluster,FAT_findFreeCluster,FAT_getNextCluster,FAT_getSectorOfCluster,PartitionAdjust,getMSDData
 EXTERN FAT_getEntryInfo,FAT_findFileInCluster,USBErrorCont,FAT_getStartingCluster,IGetKey,LBAMinus1,FAT_findEmptyEntry
 EXTERN DispHexHL,LBAPlus1,IsFinalCluster,IsFAT32,DispHexA,IsFAT12,FAT_WriteClusterDE,IPutS

DOS_countFilesInDir:
	;Counts the directory entries in a directory
	;Input:	HL is ASCIIZ string of path to directory
	;Output:	BC is the number of entries
       call FAT_getStartingCluster
       ld bc,-1
       ld de,0
$$:    push de
       push hl
       call DOS_getNextFile
       pop hl
       pop de
       inc de
       jr nc,$B
       dec de
       ld b,d
       ld c,e
       ret

DOS_getDirEntry:
	;Input:	BC points to the current entry number of a file
	;		HL points to the starting cluster of the directory
	;Output:	HL points to the directory entry of the file
       ;Returns C if not found or problems
	;Destroys OP6
       dec bc ;fall through to DOS_getNextFile

DOS_getNextFile:
	;Returns the next file in a directory
	;Input:	HL is the starting cluster of the directory
	;		BC is the current entry number of a file (pass -1 on first call to this function)
	;Output:	BC is the entry of the next file, or C if there are no more
	;		HL points to the directory entry of the next file
	;Destroys OP6
       call getMSDPtr
       ld a,h
       or l
       ld a,(ix+MSDSPC)
       ld (SPC_temp),a
       jr nz,$F
 IF FAT32_ON = 1
       call IsFAT32
       jr nz,$F
 ENDIF
       ld a,(ix+MSDRootDirSectors)
 IF FAT12_ON = 1
       call IsFAT12
       jr z,$F
       ld a,16 ;ugly FAT12 hack, assumes RootDirSectors would have been 14 (most (all?) format programs do this)
 ENDIF
$$:    ld (ix+MSDSPC),a
       inc bc
       ld (OP6+8),bc
       push hl
       call FAT_getEntryInfo
       pop hl
       ld bc,(OP6+4)
       ld a,b
       or c
       jr z,$F
DOS_GNF_C_1:
	push bc
	call FAT_getNextCluster
       call IsFinalCluster
       pop bc
       jr c,DOS_GNF_NF ;safety code
	dec bc
	ld a,b
	or c
	jr nz,DOS_GNF_C_1
$$:    ld (sector_temp),hl ;actually cluster_temp
	call FAT_getSectorOfCluster
	ld a,(OP6+3)
	ld c,a
	ld b,0
	add hl,bc
	jr nc,$F
	inc de
$$:    ;DEHL is sector...get byte offset
	call PartitionAdjust
DOS_GNF_SectorLoop:
	call getSectorBuffer
	call UFI_Read
       push de
	ld de,(OP6+1)
	call getMSDPtr
	ld l,(ix+MSDBPS)
	ld h,(ix+MSDBPS+1)
	or a
	sbc hl,de
	;The remainder
	ld b,h
	ld c,l
	pop hl
	add hl,de
	srl b
	rr c
	srl b
	rr c
	srl b
	rr c
	srl b
	rr c
	srl b
	rr c
	ld b,c
DOS_GNF_Loop:
	ld a,(hl)
       or a
	jr z,DOS_GNF_NF
       ;cp 5 ;no need to ignore Kanji character set
       ;jr z,DOS_GNF_SE
       cp 0E5h
	jr z,DOS_GNF_SE
	;found an entry?
	push hl
	ld de,11
	add hl,de
	ld a,(hl)
	and 8 ;volume label? (hope not)
	pop hl
	jr z,DOS_GNF_Ret
DOS_GNF_SE:
	ld de,(OP6+8)
	inc de
	ld (OP6+8),de
	;loop back for more
	ld de,32
	add hl,de
	djnz DOS_GNF_Loop
	;new sector or new cluster?
	call getMSDPtr
	ld a,(ix+MSDSPC)
	ld b,a
	ld a,(OP6+3)
	inc a
	cp b
	jr z,DOS_GNF_NC ;new cluster
	ld (OP6+3),a
	call LBAPlus1
       ld hl,0
	ld (OP6+1),hl
	jr DOS_GNF_SectorLoop
DOS_GNF_NC:
	xor a
	ld (OP6+1),a
	ld (OP6+2),a
	ld (OP6+3),a
	;new cluster
	ld hl,(sector_temp)
	call FAT_getNextCluster
	ld (sector_temp),hl
       call IsFinalCluster
	jr c,DOS_GNF_NF ;there are no more clusters, so it's not here
	call FAT_getSectorOfCluster
	call PartitionAdjust
	jr DOS_GNF_SectorLoop
DOS_GNF_NF:
	scf
DOS_GNF_Ret:
	push af
	ld bc,(OP6+8)
	ld a,(SPC_temp)
	call getMSDPtr
	ld (ix+MSDSPC),a
	pop af
	ret

DOS_getFileSize:
	;Gets size for a file in bytes
	;Input:      HL points to directory entry for file
	;Output:     DEHL is size (32-bit)
	;Returns C if not found
	;Always returns zero for a directory
	ld de,28
	add hl,de
	ld e,(hl)
	inc hl
	ld d,(hl)
	inc hl
	ld a,(hl)
	inc hl
	ld h,(hl)
	ld l,a
	ex de,hl
	ret

DOS_createFile:
       ;Creates new file in specified directory with size 0
	;Input:	BC points to the starting cluster of a directory
	;		HL points to an ASCIIZ string
	;Output:	File is created with size 0
	;C set if problems
       ;This function needs help, big time
	call getMSDPtr
	push bc
	ld a,b
	or c
	ld a,(ix+MSDSPC)
	ld (SPC_temp),a
	jr nz,$F
 IF FAT32_ON = 1
       call IsFAT32
       jr nz,$F
 ENDIF
	ld a,(ix+MSDRootDirSectors)
 IF FAT12_ON = 1
       call IsFAT12
       jr z,$F
       ld a,16 ;ugly FAT12 hack, see above
 ENDIF
$$:    ld (ix+MSDSPC),a
	ld de,str_temp
	call FAT_nameConvertTo11
	pop hl
       call FAT_findEmptyEntry
       jr c,DOS_CF_Done
	ex de,hl
	ld hl,str_temp
	ld bc,11
	ldir
	ex de,hl
	;attributes
	ld (hl),0
	inc hl
	;reserved
	ld (hl),0
	inc hl
	;create time, fine resolution (don't care)
	inc hl
	;create time
	call FAT_getTimeWord
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;create date
	call FAT_getDateWord
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;last access date
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;high cluster word
	ld (hl),0
	inc hl
	ld (hl),0
	inc hl
	;last modified time
	call FAT_getTimeWord
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;last modified date
	call FAT_getDateWord
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;first cluster
	ld (hl),0
	inc hl
	ld (hl),0
	inc hl
	;file size
	ld (hl),0
	inc hl
	ld (hl),0
	inc hl
	ld (hl),0
	inc hl
	ld (hl),0
       ld de,(cachetemp)
	call UFI_Write
DOS_CF_Done:
       push af
       call getMSDPtr
       ld a,(SPC_temp)
       ld (ix+MSDSPC),a
       pop af
       ret

DOS_createDirectory:
	;Creates empty directory at specified location
	;Input:	HL points to starting cluster of a directory
	;		DE points to an ASCIIZ directory name
	;Output:	New directory is created
	;		HL contains starting cluster of this directory
	;C set if problems
	call getMSDPtr
	ld a,h
	or l
	ld a,(ix+MSDSPC)
	ld (SPC_temp),a
	jr nz,$F
 IF FAT32_ON = 1
       call IsFAT32
       jr nz,$F
 ENDIF
	ld a,(ix+MSDRootDirSectors)
 IF FAT12_ON = 1
       call IsFAT12
       jr z,$F
       ld a,16 ;ugly FAT12 hack, see above
 ENDIF
$$:    ld (ix+MSDSPC),a
       ld (OP1_temp+2),hl ;used as temporary original cluster
       ex de,hl
       ld de,str_temp
       call FAT_nameConvertTo11
	call FAT_findFreeCluster
	ret c
	ld (OP1_temp),hl ;used as temporary new cluster
       ld hl,(OP1_temp+2)
       call FAT_findEmptyEntry
	ex de,hl
	ld hl,str_temp
	ld bc,11
	ldir
	ex de,hl
	;HL is offset in sector buffer just after 8.3 name (attributes)
	ld (hl),10h
	inc hl
	;reserved
	ld (hl),0
	inc hl
	;create time, fine resolution (don't care)
	inc hl
	;create time
	call FAT_getTimeWord
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;create date
	call FAT_getDateWord
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;last access date
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;high cluster word
	ld (hl),0
	inc hl
	ld (hl),0
	inc hl
	;last modified time
	call FAT_getTimeWord
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;last modified date
	call FAT_getDateWord
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;first cluster
	ld de,(OP1_temp)
	ld (hl),e
	inc hl
	ld (hl),d
	;file size
       ld b,4
$$:    inc hl
	ld (hl),0
       djnz $B
	call getMSDPtr
	ld a,(SPC_temp)
	ld (ix+MSDSPC),a
       ld de,(cachetemp)
	call UFI_Write
	ld hl,(OP1_temp)
	call FAT_getNextCluster
	ld a,0FFh
	ld (de),a
	inc de
	ld (de),a
       ld de,(cachetemp)
	call UFI_Write
	ld hl,(OP1_temp)
	call FAT_getSectorOfCluster
	call PartitionAdjust
	;LBA points to new cluster to hold '.' and '..' entries
	call getSectorBuffer
	push de
	ex de,hl
	call getMSDPtr
	ld c,(ix+MSDBPS)
	ld b,(ix+MSDBPS+1)
	B_CALL MemClear
	pop de
	ld hl,sDotEntry
	ld bc,14
	ldir
	ld hl,14
	add hl,de
	;create time
	call FAT_getTimeWord
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;create date
	call FAT_getDateWord
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;last access date
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;high cluster word (don't care)
	ld (hl),0
	inc hl
	ld (hl),0
	inc hl
	;last modified time
	call FAT_getTimeWord
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;last modified date
	call FAT_getDateWord
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;starting cluster
	ld de,(OP1_temp)
	ld (hl),e
	inc hl
	ld (hl),d
	ld b,4
$$:	inc hl
	ld (hl),0
	djnz $B
	call getSectorBuffer
	ld hl,32
	add hl,de
	ex de,hl
	push de
	ld hl,sDotDotEntry
	ld bc,14
	ldir
	pop de
	ld hl,14
	add hl,de
	;create time
	call FAT_getTimeWord
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;create date
	call FAT_getDateWord
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;last access date
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;high cluster word (don't care)
	ld (hl),0
	inc hl
	ld (hl),0
	inc hl
	;last modified time
	call FAT_getTimeWord
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;last modified date
	call FAT_getDateWord
	ld (hl),e
	inc hl
	ld (hl),d
	inc hl
	;starting cluster, DE contains original starting cluster
       ld de,(OP1_temp+2)
	ld (hl),e
	inc hl
	ld (hl),d
	ld b,4
$$:	inc hl
	ld (hl),0
	djnz $B
	call getSectorBuffer ;sector created earlier, no need for cache
	jr UFI_Write
sDotEntry:
	DB ".          ",10h,0,0
sDotDotEntry:
	DB "..         ",10h,0,0

DOS_renameFileEntry:
	;Renames a file or folder
	;Input:      HL points to an ASCIIZ string of the old file/directory name (including path) (e.g. "/folder1/old.txt")
	;            DE points to an ASCIIZ string of the new name (e.g. "new.txt")
	;Output:     Entry is renamed
	;Returns C if it failed
	push de
	call FAT_lookupPath ;this sets up (MSDLBA) for sector contaning directory entry
	ex de,hl
	pop hl
	ret c
       push hl ;move LBA back one because of FAT_lookupPath (really FAT_findEntry)
	call LBAMinus1
       pop hl
	;HL is ASCIIZ new filename
	;DE is pointer in sector buffer to directory entry
	push de
	ld de,temp_file
	call FAT_nameConvertTo11
	pop de
	;HL is 11-character new filename
	;DE is pointer in sector buffer to directory entry
	ld bc,11
	ldir
       ld de,(cachetemp)
	jr UFI_Write ;save it

DOS_deleteFileEntry:
	;Input:	HL points to the starting cluster of a directory
	;		DE points to a filename (11-character format)
	;Output:	Entry is deleted
	;If directory, entire contents are also deleted
       ;C set if problems
       ;This function needs a major overhaul in code and logic
	call FAT_findFileInCluster
	ret c
	;HL points to directory entry to delete
	push hl
	ld de,11
	add hl,de
	ld a,(hl)
	and 10000b
	pop hl
	jr z,DOS_DFE_File
	ld a,(hl)
	cp '.'
	jr nz,DOS_DFE_GoodDir
	inc hl
	ld a,(hl)
	dec hl
	cp ' '
	jr z,DOS_DFE_DotDot
	cp '.'
	jr nz,DOS_DFE_GoodDir
DOS_DFE_DotDot:
	ld (hl),0E5h
       call LBAMinus1
;	call getSectorBuffer                    ;***CACHE TEST
       ld     de,(cachetemp)                     ;***CACHE TEST
	call UFI_Write
	jr DOS_DFE_Ret
DOS_DFE_GoodDir:
	push hl
	call DOS_DFE_DotDot
	pop hl
	ld de,1Ah
	add hl,de
	ld e,(hl)
	inc hl
	ld d,(hl)
	ex de,hl
	;HL is starting cluster of directory to delete
DOS_DFE_DirLoop:
	ld bc,-1
	push hl
	call DOS_getNextFile
	ex de,hl
	pop hl
	jr c,deleteLoop
	push hl
	call DOS_deleteFileEntry ;recurse through files/directories
	pop hl
	jr DOS_DFE_DirLoop
DOS_DFE_File:
	ld (hl),0E5h
	ld de,1Ah
	add hl,de
	ld e,(hl)
	inc hl
	ld d,(hl)
	push de
       call LBAMinus1
;	call getSectorBuffer                           ;***CACHE TEST
       ld     de,(cachetemp)
	call UFI_Write
	pop hl
	ld a,h
	or l
	jr z,DOS_DFE_Ret
deleteLoop:
;	call getSectorBuffer                           ;***CACHE TEST
       ld     de,(cachetemp)                     ;***CACHE TEST
	call FAT_getNextCluster
	xor a
	ld (de),a
	inc de
	ld (de),a
 IF FAT32_ON = 1
       call IsFAT32
       jr z,$F
       inc de
       ld (de),a
       inc de
       ld (de),a
 ENDIF
$$:	call DOS_DFE_IsSameSector
;	call getSectorBuffer                    ;***CACHE TEST
       ld     de,(cachetemp)
	push hl
	call UFI_Write
	pop hl
       call IsFinalCluster
	jr nc,deleteLoop
DOS_DFE_Ret:
	or a
	ret
DOS_DFE_IsSameSector:
       call IsFinalCluster
	ret c
	push hl
	call getMSDPtr
	ld e,(ix+MSDBPS)
	ld d,(ix+MSDBPS+1)
	xor a
	add hl,hl		;AHL = 2HL
	adc a,a
 IF FAT32_ON = 1
       call IsFAT32
       jr z,$F
       add hl,hl
       adc a,a
 ENDIF
$$:	dec de
	push af
	ld a,h
	and d
	ld b,a
	ld a,l
	and e
	ld c,a		;BC = (2AHL) MOD DE
	pop af
DOS_DFE_IsSameSector_1:
	sra a
	rr h
	rr l
	srl d
	rr e
	jr nz,DOS_DFE_IsSameSector_1
	;HL is the sector...BC is the offset
	ld e,(ix+MSDFAT_sector)
	ld d,(ix+MSDFAT_sector+1)
	B_CALL cphlde
	jr z,DOS_DFE_Same
	pop hl
	ret
DOS_DFE_Same:
	pop hl
	push de
;	call getSectorBuffer                    ;***CACHE TEST
       ld     de,(cachetemp)                     ;***CACHE TEST
	ex de,hl
	pop de
	add hl,bc
	ld e,(hl)
	inc hl
	ld d,(hl)
 IF FAT32_ON = 1
       call IsFAT32
       jr z,$F
       inc hl
       ld (hl),0
       inc hl
       ld (hl),0
       dec hl
       dec hl
 ENDIF
$$:	ld (hl),0
	dec hl
	ld (hl),0
	ex de,hl
	jr DOS_DFE_IsSameSector

DOS_openFileSC:
       ;Opens a file for reading or writing by populating a file handle containing important file information (length fHandleLen)
       ;Input:      HL points to an ASCIIZ string containing a filename (NOT including path)
       ;            DE points to a location to store the file handle
       ;            BC is starting cluster of directory containing file
       ;Output:     HL points to file handle
       ;Returns C if failure
       ;This is DOS_openFile "overloaded"
	push de
	push bc
	ld de,str_temp
	call FAT_nameConvertTo11
	ex de,hl
	pop hl
	;HL needs to be cluster number, DE 11-char string
	call FAT_findFileInCluster
	pop de
	jr DOS_openFile_Cont

DOS_openFile:
       ;Opens a file for reading or writing by populating a file handle containing important file information (length fHandleLen)
       ;Input:      HL points to an ASCIIZ string containing a filename (including path)
       ;            DE points to a location to store the file handle
       ;Output:     HL points to file handle
	;Returns C if failure
	push de
	call FAT_lookupPath
	pop de
DOS_openFile_Cont:
       ;at this point, C set if not found
       ;HL is directory entry
       ;DE is location to store file handle
	ret c
	push de
	push hl
       ld de,(cachetemp)
	sbc hl,de
	srl h
	rr l
	srl h
	rr l
	srl h
	rr l
	srl h
	rr l
	srl h
	rr l
	ld a,l
	ld (temp_byte),a ;temporarily store directory entry offset
	pop hl
       pop de
	;check if directory instead of file
	ld bc,11
	add hl,bc
	ld a,(hl)
	and 18h
	ld a,errMSDNotFile
	jr nz,USBErrorCont ;either a volume label or directory
       push de
	ld bc,28-11
	add hl,bc
	ld bc,4
	ldir
	ld bc,-6
	add hl,bc
	ldi
	ldi
	;initialize seek to 0
	ld h,d
	ld l,e
	ld (hl),0
	inc de
	ld bc,6
	ldir
	pop hl
	push hl
	push hl
	pop ix
	ld (ixSave),ix
       call LBAMinus1
	ld a,(ix+MSDLBA)
	ld l,a
	ld a,(ix+MSDLBA+1)
	ld h,a
	ld a,(ix+MSDLBA+2)
	ld e,a
	ld a,(ix+MSDLBA+3)
	ld d,a
	ld ix,(ixSave)
	ld (ix+fHandDirEntryAddress),l
	ld (ix+fHandDirEntryAddress+1),h
	ld (ix+fHandDirEntryAddress+2),e
	ld (ix+fHandDirEntryAddress+3),d
	ld a,(temp_byte)
	ld (ix+fHandDirEntryOffset),a
	ld (ix+fHandFlags),0
	push ix
	call DOS_OptimizeFH
	pop ix
	pop hl
	or a
	ret

DOS_fileSeek:
       ;Positions the seek pointer in a file
       ;Input:       BC is a pointer to the file handle of an opened file
       ;             DEHL is the offset to move the seek pointer to
       ;Output:      Seek pointer is modified
       push bc
       pop ix
       ld (ix+fHandSeekOffset),l
       ld (ix+fHandSeekOffset+1),h
       ld (ix+fHandSeekOffset+2),e
       ld (ix+fHandSeekOffset+3),d
       call DOS_OptimizeFH_Sub
       or a
       ret

DOS_OptimizeFH:
	;Routine to adjust seek pointers in a file handle whenever the seek position is changed
	;Input:      IX points to start of file handle
	;Output:     File handle changed
       ;Notes:      This function looks at (ix+fHandSeekOffset) and updates the rest of the file handle
       ld l,(ix+fHandSeekOffset)
	ld h,(ix+fHandSeekOffset+1)
	ld e,(ix+fHandSeekOffset+2)
	ld d,(ix+fHandSeekOffset+3)
DOS_OptimizeFH_Sub:
	;DEHL is the 32-bit offset of the file
	;find the byte offset
	ld (ixSave),ix
	call getMSDPtr
	ld c,(ix+MSDBPS)
	ld b,(ix+MSDBPS+1)
	dec bc
	ld a,b
	and h
	ld ix,(ixSave)
	ld (ix+fHandByteOffset+1),a
	ld a,c
	and l
	ld (ix+fHandByteOffset),a
	;divide by BPS
DOS_OFH1:
	srl d
	rr e
	rr h
	rr l
	srl b
	rr c
       ld a,b ;for safety
       or c
	jr nz,DOS_OFH1
	;DEHL is now a sector offset
	call getMSDPtr
	ld a,(ix+MSDSPC)
       dec a
       push af
	ld ix,(ixSave)
	and l
	ld (ix+fHandCurSectorOffset),a
       pop af
       ;divide by SPC
DOS_OFH2:
	srl d
	rr e
	rr h
	rr l
       srl a
	jr nz,DOS_OFH2
       ld (ix+fHandCurClusterOffset),l
	ld (ix+fHandCurClusterOffset+1),h
	ld b,h
	ld c,l
       ;BC is the cluster offset
	ld l,(ix+fHandStartingCluster)
	ld h,(ix+fHandStartingCluster+1)
	ld a,b
	or c
	jr z,$F
DOS_OFH3:
	;loop through
	push bc
	push ix
	call FAT_getNextCluster
	pop ix
	pop bc
	dec bc
	ld a,b
	or c
	jr nz,DOS_OFH3
$$:	ld (ix+fHandCurCluster),l
	ld (ix+fHandCurCluster+1),h
	;look up the sector for it
	push ix
	call FAT_getSectorOfCluster
	pop ix
	;DEHL is the sector
	;add the sector offset
	ld c,(ix+fHandCurSectorOffset)
	ld b,0
	add hl,bc
	ld bc,0
	ex de,hl
	adc hl,bc
	ex de,hl
	;adjust for partition offset
	call getMSDPtr
	ld c,(ix+MSDPartStart)
	ld b,(ix+MSDPartStart+1)
	add hl,bc
	ex de,hl
	ld c,(ix+MSDPartStart+2)
	ld b,(ix+MSDPartStart+3)
	adc hl,bc
	ex de,hl
	ld ix,(ixSave)
	ld (ix+fHandSeekAddress),l
	ld (ix+fHandSeekAddress+1),h
	ld (ix+fHandSeekAddress+2),e
	ld (ix+fHandSeekAddress+3),d
       or a
	ret

DOS_fileRead:
	;Reads bytes from a file
	;Input:	HL points to file handle of opened file
	;		DE is where to store bytes read
	;		BC is the number of bytes to read
	;Output:	Bytes are read to DE
	;		BC is the number of bytes actually read
	;Returns C if failure or EOF encountered
	push hl
	pop ix
	ld (ixSave),ix
	ld (ix+fHandFlags),0
	ld hl,0
	ld (bytes_read),hl
	ld (store_ptr),de
	;IX is the pointer to the file handle
	;Check how many bytes we are reading...is it too much?
	ld a,(ix+0)
	sub (ix+9)		;Subtract position in file to get how many
	ld l,a			;bytes left to read
	ld a,(ix+1)
	sbc (ix+10)
	ld h,a
	ld a,(ix+2)
	sbc (ix+11)
	ld e,a
	ld a,(ix+3)
	sbc (ix+12)
	ld d,a
	jr nc,skipifoverread
	ld hl,0
	ld de,0 ;if we are over, then load 0 bytes left to read
skipifoverread:
	push bc
	or a
	push hl ;preserve dehl
	sbc hl,bc
	pop hl
	ex de,hl
	ld bc,0
	push hl
	sbc hl,bc
	pop hl
	ex de,hl
	pop bc
	jr nc,$F
	ld c,l	;bytes to read
	ld b,h
	ld a,flagEOF
	or (ix+21)
	ld (ix+21),a
$$:	ld (remaining),bc
	ld a,b
	or c ;make sure there is something to read.
	jr z,dontfread
DOS_FR_Loop:
	;Shall we read some bytes? Sure!
	ld bc,(remaining)
	call getMSDPtr
	ld l,(ix+MSDBPS)
	ld h,(ix+MSDBPS+1)
	ld ix,(ixSave)
	ld e,(ix+13)
	ld d,(ix+14)
	or a
	sbc hl,de
	;HL is the most bytes we can read at once
	sbc hl,bc
	;If BC > HL, we have a problem
	jr nc,$F
	add hl,bc
	ld b,h
	ld c,l ;We will only read BC
$$:
	;BC is the number of bytes to read
	push bc
	push de
	ld l,(ix+15)
	ld h,(ix+16)
	call getMSDPtr
	ld (ix+MSDLBA),l
	ld (ix+MSDLBA+1),h
	ld ix,(ixSave)
	ld l,(ix+17)
	ld h,(ix+18)
	call getMSDPtr
	ld (ix+MSDLBA+2),l
	ld (ix+MSDLBA+3),h
	ld ix,(ixSave)
	call getSectorBuffer
	push ix
	call UFI_Read
	pop ix
;	call getSectorBuffer                    ;***CACHE TEST
	ex de,hl
	pop de
	add hl,de
	;HL is the offset...now copy BC (on stack) bytes
	pop bc
	push bc
	ld de,(store_ptr)
	ldir
	ld (store_ptr),de ;new location to store the next few bytes
	pop bc
	ld hl,(bytes_read)
	add hl,bc
	ld (bytes_read),hl
	ld hl,(remaining)
	or a
	sbc hl,bc
	ld (remaining),hl
	;Now we update pointers and things
	ld l,(ix+9)
	ld h,(ix+10)
	ld e,(ix+11)
	ld d,(ix+12)
	add hl,bc
	jr nc,$F
	inc de
$$:
	ld (ix+9),l
	ld (ix+10),h
	ld (ix+11),e
	ld (ix+12),d
	ld l,(ix+13)
	ld h,(ix+14)
	add hl,bc
	ld (ix+13),l
	ld (ix+14),h
	;Are we at a sector boundary?
	call getMSDPtr
	ld c,(ix+MSDBPS)
	ld b,(ix+MSDBPS+1)
	ld ix,(ixSave)
	dec bc
	ld a,h
	and b
	ld b,a
	ld a,l
	and c
	or b		;Z if we are
	jr nz,DOS_FR_PreLoop
	ld (ix+13),0
	ld (ix+14),0
	;Increment sector counts
	inc (ix+15)
	jr nz,DOS_FR_2
	inc (ix+16)
	jr nz,DOS_FR_2
	inc (ix+17)
	jr nz,DOS_FR_2
	inc (ix+18)
DOS_FR_2:
	call getMSDPtr
	ld a,(ix+MSDSPC)
	ld ix,(ixSave)
	ld b,a
	ld a,(ix+8)
	inc a
	cp b
	jr z,DOS_FR_NextCluster
	ld (ix+8),a
DOS_FR_PreLoop:
	ld bc,(remaining)
	ld a,b
	or c
	jr nz,DOS_FR_Loop
dontfread:
	ld bc,(bytes_read)
	ld a,(ix+21)
	and flagEOF
	ret z
	scf
	ret
DOS_FR_NextCluster:
	xor a
	ld (ix+8),a
	inc (ix+6)
	jr nz,$F
	inc (ix+7)
$$:
	ld l,(ix+19)
	ld h,(ix+20)
	push ix
	call FAT_getNextCluster
	pop ix
	push ix
	ld (ix+19),l
	ld (ix+20),h
	call FAT_getSectorOfCluster
	call getMSDPtr
	ld c,(ix+MSDPartStart)
	ld b,(ix+MSDPartStart+2)
	add hl,bc
	ex de,hl
	ld c,(ix+MSDPartStart+2)
	ld b,(ix+MSDPartStart+3)
	adc hl,bc
	ex de,hl
	pop ix
	ld (ix+15),l
	ld (ix+16),h
	ld (ix+17),e
	ld (ix+18),d
	jr DOS_FR_PreLoop

DOS_fileWrite:
	;Input:	HL is pointer to file handle of opened file
	;		DE is location of bytes to write
	;		BC is the number of bytes to write
	;Output:	Bytes are written and file is grown if necessary
	;		BC is the number of bytes actually written
	;Returns C if it failed
	push hl
	pop ix
	ld (ixSave),ix
	ld hl,0
	ld (bytes_read),hl ;actually bytes written :)
	ld (store_ptr),de
	ld (remaining),bc
	ld l,(ix+fHandStartingCluster)
	ld h,(ix+fHandStartingCluster+1)
	ld a,h
	or l
	jr nz,DOS_FW_Loop ;space already exists
       ;No space has been allocated for file so far
       ld hl,(ixSave)
       ld b,19
$$:    ld (hl),0 ;zero out relevant parts of file handle
       inc hl
       djnz $B
	call FAT_findFreeCluster
       ret c
       ld ix,(ixSave)
	ld (ix+fHandStartingCluster),l ;save allocated cluster
	ld (ix+fHandStartingCluster+1),h
	ld (ix+fHandCurCluster),l
	ld (ix+fHandCurCluster+1),h
 IF FAT12_ON = 1
       call IsFAT12
       jr z,$F
       ld bc,0FFFFh
       push hl
       call FAT_WriteClusterDE
       pop hl
       jr DOS_FW_Written
 ENDIF
$$:	ld a,0FFh
	ld (de),a
	inc de
	ld (de),a
       ld de,(cachetemp)
	push hl
	call UFI_Write
	pop hl
DOS_FW_Written:
	call FAT_getSectorOfCluster
	call PartitionAdjust
	call getMSDPtr
	ld l,(ix+MSDLBA)
	ld h,(ix+MSDLBA+1)
	ld ix,(ixSave)
	ld (ix+fHandSeekAddress),l
	ld (ix+fHandSeekAddress+1),h
	call getMSDPtr
	ld l,(ix+MSDLBA+2)
	ld h,(ix+MSDLBA+3)
	ld ix,(ixSave)
	ld (ix+fHandSeekAddress+2),l
	ld (ix+fHandSeekAddress+3),h
DOS_FW_Loop:
	ld bc,(remaining)
	call getMSDPtr
	ld l,(ix+MSDBPS)
	ld h,(ix+MSDBPS+1)
	ld ix,(ixSave)
	ld e,(ix+fHandByteOffset)
	ld d,(ix+fHandByteOffset+1)
	or a
	sbc hl,de
	;HL is the most bytes we can write at once (for the current sector)
	sbc hl,bc
	;if BC > HL, we have a problem
	jr nc,$F
	add hl,bc
	ld b,h
	ld c,l ;only write HL then (for the current sector)
$$:    push bc
	push de
	ld l,(ix+fHandSeekAddress)
	ld h,(ix+fHandSeekAddress+1)
	call getMSDPtr
	ld (ix+MSDLBA),l
	ld (ix+MSDLBA+1),h
	ld ix,(ixSave)
	ld l,(ix+fHandSeekAddress+2)
	ld h,(ix+fHandSeekAddress+3)
	call getMSDPtr
	ld (ix+MSDLBA+2),l
	ld (ix+MSDLBA+3),h
	call getSectorBuffer
	call UFI_Read
	pop hl
	add hl,de
	pop bc
	push bc
	ex de,hl
	ld hl,(store_ptr)
	ldir
	ld (store_ptr),hl
	ex de,hl
	pop bc
	ld hl,(bytes_read)
	add hl,bc
	ld (bytes_read),hl
	ld hl,(remaining)
	or a
	sbc hl,bc
	ld (remaining),hl
       ld ix,(ixSave)
	ld l,(ix+fHandSeekOffset)
	ld h,(ix+fHandSeekOffset+1)
	ld e,(ix+fHandSeekOffset+2)
	ld d,(ix+fHandSeekOffset+3)
	add hl,bc
	jr nc,$F
	inc de
$$:    ld (ix+fHandSeekOffset),l
	ld (ix+fHandSeekOffset+1),h
	ld (ix+fHandSeekOffset+2),e
	ld (ix+fHandSeekOffset+3),d
	ld l,(ix+fHandByteOffset)
	ld h,(ix+fHandByteOffset+1)
	add hl,bc
	ld (ix+fHandByteOffset),l
	ld (ix+fHandByteOffset+1),h
	push hl
       ld de,(cachetemp)
	call UFI_Write
	pop hl
	;at sector boundary?
	call getMSDPtr
	ld c,(ix+MSDBPS)
	ld b,(ix+MSDBPS+1)
	ld ix,(ixSave)
	dec bc
	ld a,h
	and b
	ld b,a
	ld a,l
	and c
	or b ;Z if so
	jr nz,DOS_FW_PreLoop
	ld (ix+fHandByteOffset),0
	ld (ix+fHandByteOffset+1),0
	;increment sector counts
	inc (ix+fHandSeekAddress)
	jr nz,$F
	inc (ix+fHandSeekAddress+1)
	jr nz,$F
	inc (ix+fHandSeekAddress+2)
	jr nz,$F
	inc (ix+fHandSeekAddress+3)
$$:    call getMSDPtr
	ld a,(ix+MSDSPC)
	ld ix,(ixSave)
	ld b,a
	ld a,(ix+fHandCurSectorOffset)
	inc a
	cp b
	jr z,DOS_FW_NextCluster
	ld (ix+fHandCurSectorOffset),a
DOS_FW_PreLoop:
	ld bc,(remaining)
	ld a,b
	or c
	jr nz,DOS_FW_Loop
;adjust data from file handle before returning
;if seek pointer > file size, overwrite file size to seek pointer
	ld c,(ix+fHandFileSize+2)
	ld b,(ix+fHandFileSize+3)
	ld l,(ix+fHandSeekOffset+2)
	ld h,(ix+fHandSeekOffset+3)
	or a
	sbc hl,bc
	jr c,$F ;file size > seek pointer, everything is okay
	ld c,(ix+fHandFileSize)
	ld b,(ix+fHandFileSize+1)
	ld l,(ix+fHandSeekOffset)
	ld h,(ix+fHandSeekOffset+1)
	sbc hl,bc
	jr c,$F ;file size > seek pointer, everything is okay
       ;seek pointer > file size, adjust directory entry
	ld l,(ix+fHandSeekOffset)
	ld h,(ix+fHandSeekOffset+1)
	ld e,(ix+fHandSeekOffset+2)
	ld d,(ix+fHandSeekOffset+3)
	ld (ix+fHandFileSize),l
	ld (ix+fHandFileSize+1),h
	ld (ix+fHandFileSize+2),e
	ld (ix+fHandFileSize+3),d
       ld hl,MSDLBA
       call getMSDData
       ex de,hl
       ld hl,(ixSave)
       ld bc,fHandDirEntryAddress
       add hl,bc
       ld bc,4
       ldir
	call getSectorBuffer
	call UFI_Read
       ld ix,(ixSave)
	ld h,0
	ld l,(ix+fHandDirEntryOffset)
	add hl,hl ;*2
	add hl,hl ;*4
	add hl,hl ;*8
	add hl,hl ;*16
	add hl,hl ;*32
	add hl,de
	ld de,1Ah
	add hl,de
	ld a,(ix+fHandStartingCluster)
	ld (hl),a
	inc hl
	ld a,(ix+fHandStartingCluster+1)
	ld (hl),a
	inc hl
	ld a,(ix+fHandFileSize)
	ld (hl),a
	inc hl
	ld a,(ix+fHandFileSize+1)
	ld (hl),a
	inc hl
	ld a,(ix+fHandFileSize+2)
	ld (hl),a
	inc hl
	ld a,(ix+fHandFileSize+3)
	ld (hl),a
       ld de,(cachetemp)
	call UFI_Write
$$:	ld bc,(bytes_read)
	or a
	ret
DOS_FW_NextCluster:
;if on final cluster, create new one
	xor a
	ld (ix+fHandCurSectorOffset),a
	inc (ix+fHandCurClusterOffset)
	jr nz,$F
	inc (ix+fHandCurClusterOffset+1)
$$:    ld l,(ix+fHandCurCluster)
	ld h,(ix+fHandCurCluster+1)
	push ix
	ld (cur_cluster),hl
	call FAT_getNextCluster
	pop ix
       call IsFinalCluster
	jr nc,DOS_FW_NoNewCluster
	call FAT_findFreeCluster
       ret c
	push hl
	ld a,0FFh
	ld (de),a
	inc de
	ld (de),a
 IF FAT32_ON = 1
       call IsFAT32
       jr z,$F
       inc de
       ld (de),a
       inc de
       ld (de),a
 ENDIF
$$:    ld de,(cachetemp)
	call UFI_Write
	ld hl,(cur_cluster)
	call FAT_getNextCluster
	pop hl
	ld a,l
	ld (de),a
	inc de
	ld a,h
	ld (de),a
 IF FAT32_ON = 1
       call IsFAT32
       jr z,$F
       inc de
       xor a
       ld (de),a
       inc de
       ld (de),a
 ENDIF
$$:	push hl
       ld de,(cachetemp)
	call UFI_Write
	pop hl
       ld ix,(ixSave)
DOS_FW_NoNewCluster:
	ld (ix+fHandCurCluster),l
	ld (ix+fHandCurCluster+1),h
       push ix
	call FAT_getSectorOfCluster
	call PartitionAdjust
	pop ix
	ld (ix+fHandSeekAddress),l
	ld (ix+fHandSeekAddress+1),h
	ld (ix+fHandSeekAddress+2),e
	ld (ix+fHandSeekAddress+3),d
	jr DOS_FW_PreLoop

DOS_updateAttributes:
       ;Updates "last modified" date/time attributes for file
       ;Input:       HL points to the file handle of opened file
       ;             B is flags byte
       ;                    Bit 0: Change "archived" status to value in C
       ;                    Bit 1: Change "read-only" status to value in C
       ;                    Bit 2: Change "hidden" status to value in C
       ;             C is byte of values to change
       ;                    Bit 0: New "archived" status
       ;                    Bit 1: New "read-only" status
       ;                    Bit 2: New "hidden" status
       ;Output:      "Last modified" date/time attributes are updated
       ;             Any other file attributes specified are updated as well
       ;Notes:       This is like closing a file, but not required
       ;Returns C set if problems
       ;get old attribute byte
       push bc
       call getMSDPtr
       ld de,fHandDirEntryAddress
       add hl,de
       ld a,(hl)
       ld (ix+MSDLBA),a
       inc hl
       ld a,(hl)
       ld (ix+MSDLBA+1),a
       inc hl
       ld a,(hl)
       ld (ix+MSDLBA+2),a
       inc hl
       ld a,(hl)
       ld (ix+MSDLBA+3),a
       inc hl
       ld l,(hl)
       ld h,0
       push hl
       call getSectorBuffer
       call UFI_Read
       ld (cachetemp),de
       pop hl
       add hl,hl ;*2
       add hl,hl ;*4
       add hl,hl ;*8
       add hl,hl ;*16
       add hl,hl ;*32
       add hl,de
       ld (address),hl
       ;update date/time
       call FAT_getTimeWord
       ld bc,16h
       add hl,bc
       ld (hl),e
       inc hl
       ld (hl),d
       inc hl
       call FAT_getDateWord
       ld (hl),e
       inc hl
       ld (hl),d
       ld hl,(address)
       ld de,11
       add hl,de
       ld (address),hl
       ld a,(hl)
       ld (checksum),a ;checksum used as temporary storage for old attribute byte
       pop bc
       ;B = flags byte, C = new values
       srl b
       jr nc,DOS_UA_NoArchive
       ld a,c
       and 00000001b
       sla a
       sla a
       sla a
       sla a
       sla a
       ld (temp_mem),a
DOS_UA_NoArchive:
       srl b
       jr nc,DOS_UA_NoReadOnly
       ld a,c
       and 00000010b
       sra a
       ld (temp_mem+1),a
DOS_UA_NoReadOnly:
       srl b
       jr nc,DOS_UA_NoHidden
       ld a,c
       and 00000100b
       sra a
       ld (temp_mem+2),a
DOS_UA_NoHidden:
       ld a,(checksum)
       and 11011100b
       ld b,a
       ld a,(temp_mem)
       or b
       ld b,a
       ld a,(temp_mem+1)
       or b
       ld b,a
       ld a,(temp_mem+2)
       or b
       ld hl,(address)
       ld (hl),a
       ld de,(cachetemp)
       jr UFI_Write
