 include "settings.inc"
  include "../ti83plus.inc"
 include "equates.inc" 
 SEGMENT Main
 GLOBALS ON
 EXTERN SendControlCmd,SetAPortA0,Send8PortA0,IPutS,IGetKey,GetAPortA0,USBSaveSP,USBInfo,DDBuf,CDBuf
 EXTERN Get8PortA0,WaitTimer2,SendOutCmdWait,SendOutCmdNoWait,IsDeviceConnected,USBHostInit,SetupInPipe
 EXTERN SetupOutPipe,DispHexA,USBReqData,CpHLDE,WaitTimer32,SetHostAddress,GetPipeInfo,SetupUSBHook
 EXTERN USBHostInitSafe,WaitTimer16,SetHubPortFeature,ConfigureHub

;USB high-level routines



SetCallBack:
;Sets driver callback address
;Input
; DE = callback address
       ld     hl,(MemPtr)
       ld     bc,offCallBackTable
       add    hl,bc
       ld     (hl),e
       inc    hl
       ld     (hl),d
       ret





Pop1USBErrorFindPipe:
       pop    hl
       jp     USBErrorFindPipe

FindPipe:
;Returns the # of the pipe matching certain characteristics
; (Doesn't check packet size.  Would you ever want this?)
;Input
; B = pipeInt or pipeBulk
; C = pipeIn or pipeOut
; HL -> descriptor buf
;Output
; NC = success
; B = pipe #

       push   bc       
       push   hl
       ld     b,descConfig
       call   ReadDescriptor
       pop    hl
       jr     c,Pop1USBErrorFindPipe
       pop    bc
FindPipeHL:
       push   bc
       ld     b,descEndpoint
       call   FindDescriptor
       jp     c,Pop1USBErrorFindPipe
       pop    bc
       call   FindPipeCheck
       ret    nc
       push   bc

FindPipeEndpointLoop:
       ld     b,descEndpoint
       call   FindDescriptorAgain
       jr     c,Pop1USBErrorFindPipe
       pop    bc
       call   FindPipeCheck
       ret    nc
       push   bc

       jr     FindPipeEndpointLoop
FindPipeCheck:
;Needs to setup the pipe based on the endpoint descriptor at HL
;Should return HL -> next descriptor
       push   de
       inc    hl
       inc    hl
       ld     d,(hl)                ;Endpoint address
       inc    hl
       ld     a,(hl)
       and    3
       add    a,a
       add    a,a
       add    a,a
       add    a,a
       ld     e,a                  ;Endpoint type
       inc    hl
       inc    hl
       inc    hl
       inc    hl
       ld     a,d
       and    80h
       rlca
       xor    1
       inc    a
       cp     c
       jr     nz,$f
       ld     a,e
       cp     b
       jr     nz,$f
       ld     a,d
       and    7Fh
       ld     b,a
       pop    de
       ret                  ;return nc
$$:
       scf
       pop    de
       ret

 COMMENT ~ OLD FindPipe

       ld     h,b           ;pipe type
       ld     l,d           ;pipe direction

       ld     b,1           ;Start looking at pipe 1
FindPipeLoop:
       push   bc
       push   hl
       call   GetPipeInfo
       DEBUG  b
       DEBUG  d
       ld     a,b           ;pipe type
       pop    hl
       pop    bc
       jp     c,USBErrorFindPipe
       cp     h
       jr     nz,FindPipeCont
       ld     a,d
       and    l
       ret    nz
FindPipeCont:
       inc    b
       cp     16
       jr     nz,FindPipeLoop
       jp     USBErrorFindPipe
 ~

SendDataPart:
;Send B Bytes at HL to endpoint C
; Don't send "end of data" to USB controller
;NC = success
       DISAVE
       call   SendDataDo
       EIRESTORE
       ret


SendData:
       DISAVE
       call   SendDataSafe
       EIRESTORE
       ret

SendDataSafe:
;Send B bytes at HL to endpoint C
;NC = success

       push   bc
       call   SendDataDo
       pop    bc
       jr     c,$f
       push   af                          ;Save flags
       ld     b,c                         ;Endpoint #
       ld     c,8
       call   SendOutCmdNoWait
       pop    af              
$$:      

       ret

SendDataDo:
       LOG    SendData,c
       ld     a,c
       out    (8Eh),a
;       xor    a
;       out    (5Bh),a
       push   hl
       ld     hl,(MemPtr)
       ld     de,offPacketSizeTable
       add    hl,de
       ld     e,c
       ld     d,0
       add    hl,de
       ld     d,(hl)                      ;D = max packet size
       ld     e,b                         ;E = bytes left to send
       pop    hl

       ld     a,c
       add    a,0A0h
       ld     c,a                         ;C = port in A0 range
SendLoop:
       ld     b,d                         ;Send max size
       ld     a,e
       sub    d                           ;Unless bytes left - max packet carries
       ld     e,a
       jr     nc,$f
;       ld     b,e
       add    a,d
       ld     b,a
       ld     e,0                         ;This is the last packet
$$:

       LOG    DataStart,b
 IF  LOGGING_ON=0
       otir
 ELSE
 $$:
       ld     a,(hl)
       LOG    Data,a
       out    (c),a
       inc    hl
;       dec    b
       djnz   $b
 ENDIF
       push   bc
       ld     a,c
       sub    0A0h
       ld     b,a
       ld     c,01


       call   SendOutCmdWait
       LOG    DataEnd
       pop    bc
       jr     c,SendDataRetC
       ld     a,e
       or     a
       jr     nz,SendLoop
;       ei
;       ld     a,1
 ;      out    (5Bh),a


       xor    a
       ret
SendDataRetC:
;       ei     
;       ld     a,1
 ;      out    (5Bh),a
       scf    
       ret


AutoSetupNoInit:
       LOG    AutoSetup
       ld     bc,ASDone
       push   bc
       DISAVE
       push   hl
       push   de
       jr     AutoSetupCont

AutoSetup:
       DISAVE
       call   AutoSetupSafe
ASDone:
       EIRESTORE
       ret

AutoSetupSafe:
;This automatically sets up the USB device if possible
;It will initialize, configure, and setup pipes in the default interface/config
;Will automatically configure up to one incoming pipes
;Input
; HL -> buffer for descriptors (strongly recommend be zero'd)
; DE -> input pipe callback routine
;Output
; NC = success
       ld     (8000h),hl
       LOG    AutoSetup       
       call   IsDeviceConnected
       jr     z,USBErrorNoDevice
       push   hl
       push   de
       call   USBHostInitSafe
       jr     c,Pop2RetC
AutoSetupCont:
       ld     hl,(MemPtr)          ;Erase call back table, so we can get descriptor #s later
       ld     de,offCallBackTable
       add    hl,de
;       ld     bc,16
;       B_CALL MemClear
       pop    de
       ld     (hl),e
       inc    hl
       ld     (hl),d               ;***** NEW CODE TO SET CALLBACK

       pop    hl
AutoSetupRedux:
       push   hl
       ld     b,descConfig
       call   ReadDescriptorSafe
       jr     c,Pop2RetC
       pop    hl
       push   hl
       ld     de,5
       add    hl,de
       ld     b,(hl)
       call   ConfigureDeviceSafe
       jr     c,Pop2RetC
       pop    hl
       ld     b,descEndpoint
       call   FindDescriptor
       jr     c,AutoSetupDone
       push   de
       call   AutoSetupPipe
       pop    de
AutoSetupEndpointLoop:
       ld     b,5
       call   FindDescriptorAgain
       jr     c,AutoSetupDone
       push   de
       call   AutoSetupPipe
       pop    de
       jr     AutoSetupEndpointLoop
AutoSetupPipe:
;Needs to setup the pipe based on the endpoint descriptor at HL
;Should return HL -> next descriptor
       inc    hl
       inc    hl
       ld     b,(hl)                ;Endpoint address
       inc    hl
       ld     a,(hl)
       and    3
       add    a,a
       add    a,a
       add    a,a
       add    a,a
       ld     c,a                  ;Endpoint type
       inc    hl
       ld     d,(hl)
       srl    d
       srl    d
       srl    d                    ;packet size
       jr     nz,$f
       inc    d
$$:
       inc    hl
       inc    hl
       ld     e,(hl)               ;Polling interval
;       di     
;       pop    af                   ;return address
;       pop    iy                   ;de val
;       pop    ix                   ;callbcak routine
;       push   ix
;       push   iy
;       push   af
;       ld     iy,flags
;       ei
       inc    hl
       push   hl
;       push   ix
;       pop    hl
       bit    7,b
       res    7,b
       jr     z,AutoSetupOutPipe
       call   SetupInPipe
       pop    hl
       ret
AutoSetupOutPipe:
       call   SetupOutPipe
       pop    hl
       ret
AutoSetupDone:
;       pop    de
 IF HUBCODE_ON=1
       ld     hl,(8000h)
       push   hl
       call   ConfigureHub
       pop    hl
       jr     nc,AutoSetupRedux
 ENDIF

       or     a
       ret

       
Pop2RetC:
       pop    hl
PopRetC:
       pop    hl
RetC:
       scf
       ret              

SetAddress:
       DISAVE
       call   SetAddressSafe
       EIRESTORE
       ret

SetAddressSafe:
;Set current control device to next available address, and make that the active address
       ld     b,30
$$:
       push   bc
       call   SetAddressSafeDo
       pop    bc
       ret    nc
       xor    a
       out    (8Eh),a
       out    (91h),a
       call   WaitTimer32                 ;Perhaps not the most elegant way to deal with NAKs
       djnz   $b
       scf
       ret

SetAddressSafeDo:
       ld     hl,(MemPtr)
       ld     de,offHighAddr
       add    hl,de
       inc    (hl)
       ld     a,(hl)               ;A = new address to allocate
       LOG    SetAddress,a
       ld     hl,SetAddrReqData
       ld     de,OutBuf
       ld     bc,8
       push   de
       ldir
       ld     (OutBuf+2),a

       pop    hl
       ld     b,0                  ;No data returned from periph
       push   af
       call   SendControlDataSafe
       pop    bc                   ;B = address
       ld     a,b
       ret    c
       call   SetHostAddress
       jp     WaitTimer2



SendControlOut:
       DISAVE
       call   SendControlOutSafe
       EIRESTORE
       ret

SendControlOutSafe:
;Input
; B = Number of bytes in data packet 
; DE = Data packet to send
; HL = Setup packet to send
;NC = success
       ld     a,b
       add    a,80h
       LOG    SendControlData,a
       push   bc
       call   Send8PortA0
       ld     a,0Ah
       call   SendControlCmd
       pop    bc
       jr     c,SendControlDataDone

 COMMENT ~ TESTING
       dec    b
       sra    b
       sra    b
       sra    b
       inc    b

       ex     de,hl
SendControlOutLoop:
       push   bc
       call   Send8PortA0

       ld     a,2
       call   SendControlCmd
       pop    bc
       jr     c,SendControlDataDone
       djnz   SendControlOutLoop
 ~
; *** TESTING
       ex     de,hl
$$:
       ld     a,(hl)
       out    (0A0h),a
       inc    hl
       djnz   $b
       ld     a,2
       call   SendControlCmd
       jr     c,SendControlDataDone
       ld     a,60h
       call   SendControlCmd          
       jr     SendControlDataDone

GetDeviceStatus:
; DE = where to put response
;NC = success
       ld     b,2
       ld     hl,GetStatusReqData

SendControlData:
       DISAVE
       call   SendControlDataSafe
       EIRESTORE
       ret
SendControlDataSafe:
;Input
; B = number of bytes expected to receive
; DE = Where to put said bytes
; HL = data to send
;NC = success
       LOG    SendControlData,b
       inc    b
       dec    b
       jr     nz,SendControlDataRec
       call   Send8PortA0
       ld     a,0Ah
       call   SendControlCmd
       jr     c,SendControlDataDone
       ld     a,60h
       call   SendControlCmd           
SendControlDataDone:
       ret
SendControlDataRec:
       push   bc
       call   Send8PortA0
       pop    bc
       ld     a,0Ah
       call   SendControlCmd
       jr     c,SendControlDataDone
       ex     de,hl


       push   hl
       ld     hl,(MemPtr)
       ld     de,offPacketSizeTable
       add    hl,de
       ld     e,(hl)                      ;E = max packet size for control pipe
       pop    hl
       ld     d,b                         ;D = total bytes to send

SendControlDataRecLoop:
       ld     a,e
       cp     b
       jr     nc,$f
       ld     b,e
$$:
       ld     a,d
       sub    b
       ld     d,a

       ld     a,20h
       call   SendControlCmd
       jr     c,SendControlDataDone

       ld     c,0A0h
       inir
       ld     b,d
       inc    d
       dec    d
       jr     nz,SendControlDataRecLoop
       ld     a,42h
       call   SendControlCmd
       jr     SendControlDataDone


ConfigureDevice:
       DISAVE
       call   ConfigureDeviceSafe
       EIRESTORE
       ret

ConfigureDeviceSafe:
;Needs to check to make sure device is initialized, then fall through below
       ld     a,b

SetConfigurationA:
;NC = success
       push   af
       ld     hl,SetConfigReqData
       ld     de,OutBuf
       ld     bc,8
       push   de
       ldir
       pop    hl
       pop    af
       ld     (OutBuf+2),a
       call   Send8PortA0
       ld     a,0Ah
       call   SendControlCmd
       ret    c
       ld     a,60h
       call   SendControlCmd
       ret    c
       xor    a
       out    (8Eh),a
       out    (91h),a
       jp     WaitTimer2                  ;Returns NC




GetMaxPacketSize:
       DISAVE
       call   GetMaxPacketSizeSafe
       EIRESTORE
       ret

GetMaxPacketSizeSafe:
       call   GetMaxPackDo
       ret    nc
       xor    a
       out    (8Eh),a
       out    (91h),a
       ld     bc,0FFFFh
$$:
       ex     hl,(sp)
       ex     hl,(sp)
       ex     hl,(sp)
       ex     hl,(sp)
       dec    bc
       ld     a,b
       or     c
       jr     nz,$b
       call   GetMaxPackDo
       ret

GetMaxPackDo:
;NC = success
       LOG    GetMaxPacketSize
       ld     hl,DDReqDataSmall
       call   Send8PortA0
       LOG    Custom,1                           ;*****TESTING
       ld     a,0Ah
       call   SendControlCmd
       ret    c
       LOG    Custom,2                           ;*****TESTING
       ld     a,20h
       call   SendControlCmd
       ret    c
       LOG    Custom,3                           ;*****TESTING
       ld     hl,OutBuf
       call   Get8PortA0
       ld     a,42h
       call   SendControlCmd
       ret    c
       LOG    Custom,4                           ;*****TESTING
       ld     a,(OutBuf+7)
       ld     hl,(MemPtr)
       ld     de,offPacketSizeTable
       add    hl,de
       ld     (hl),a
       ret


GetIDs:
       DISAVE
       call   GetIDsSafe
       EIRESTORE
       ret

GetIDsSafe:
;HL-> temporary buffer for device descriptor
;Returns 
; BC = version
; DE = product ID
; HL = vendor ID
; NC = success
       LOG    GetIDs
        push   hl
       ld     b,descDevice
       call   ReadDescriptorSafe
       pop    hl
       jp     c,USBErrorReadDescriptor
       ld     de,13
       add    hl,de
       ld     b,(hl)
       dec    hl
       ld     c,(hl)
       dec    hl
       ld     d,(hl)
       dec    hl
       ld     e,(hl)
       dec    hl
       ld     a,(hl)
       dec    hl
       ld     l,(hl)
       ld     h,a
       ret


GetClass:
       DISAVE
       call   GetClassSafe
       EIRESTORE
       ret

GetClassSafe:
;Read the class from the device descriptor, or the default config/interface if
; the device descriptor points to them.
;HL -> buffer to load descriptor (suggest you clear the buffer before calling routine)
;Output
; A = class
; B = subclass
; C = protocol
; NC = success
       LOG    GetClass
       push   hl
       ld     b,descDevice
       call   ReadDescriptorSafe
       pop    hl
       ret    c
       ld     de,4
       add    hl,de
       ld     a,(hl)
       or     a
       jr     z,GetClassInterface
GetClassCont
       inc    hl
       ld     b,(hl)
       inc    hl
       ld     c,(hl)
       ret
GetClassInterface:
       sbc    hl,de
       push   hl
       ld     b,descConfig
       call   ReadDescriptorSafe
       pop    hl
       ret    c
       ld     b,4
       call   FindDescriptor
       ret    c
       ld     de,5
       add    hl,de
       ld     a,(hl)
       jr     GetClassCont

FindDescriptor:
;Find the first instance of a specific descriptor the config descriptor
;Suggest that the descriptor list be zero terminated
;Input
; HL -> descriptor list
; B = descriptor type
;Output
; NC = success
; HL -> descriptor
; DE -> after descriptors
       push   hl
       inc    hl
       inc    hl
       ld     e,(hl)
       inc    hl
       ld     d,(hl)
       pop    hl
       push   hl
       add    hl,de
       ex     de,hl                ;DE -> address after end of config desc data
       pop    hl
FindDescriptorAgain:
;DE -> end of descriptors
       ld     c,b
       ld     b,50
$$:
       ld     a,(hl)
       or     a
       jr     z,USBErrorFindDescriptor
       push   de
       ld     e,a
       inc    hl
       ld     a,(hl)
       dec    hl
       cp     c
       jr     z,FindDescDone
       ld     d,0
       add    hl,de
       pop    de
       call   CpHLDE
       jr     z,RetC
       djnz   $b
       jp     USBErrorFindDescriptor
FindDescDone:
       pop    de
       ret

ReadDescriptor:
;Make sure device is initialized
;Don't call from interrupt
;HL = where to load descriptor
;B = descriptor type 
;C = index for string descriptor or size for report des
;NC = success
       DISAVE
       call   ReadDescriptorSafe
       EIRESTORE
       ret

ReadDescriptorSafe:
       LOG    ReadDescriptor,b
       ex     hl,de                
       ld     a,b
       cp     descDevice
       jr     z,ReqDeviceDes
       cp     descConfig
       jr     z,ReqConfigDes
       cp     descHub
       jr     z,ReqHubDes


       ex     de,hl         ;Stunningly confusing
       cp     descReport
       jr     z,ReqReportDes

       cp     descString
       ld     a,c
       jr     z,ReqStringDes
       jp     USBErrorEPGetDes

ReqReportDes:
       call   Send8PortA0
       ld     a,0Ah
       call   SendControlCmd
       ret    c
       ld     a,20h
       call   SendControlCmd
       ret    c
       ex     de,hl
       call   GetReportDescriptor
       ret    c
       ld     a,42h
       jp     SendControlCmd




ReqConfigDes:
       ld     hl,CDReqData
       call   Send8PortA0
       ld     a,0Ah
       call   SendControlCmd
       ret    c
       ld     a,20h
       call   SendControlCmd
       ret    c
       ex     de,hl
       call   GetConfigDescriptor
       ret    c
       ld     a,42h
       jp     SendControlCmd




ReqHubDes:
       ld     hl,HubDReqData
       jr     ReqDeviceDesCont

ReqDeviceDes:
       ld     hl,DDReqData
ReqDeviceDesCont:
       call   Send8PortA0

       ld     a,0Ah
       call   SendControlCmd
       ret    c
;       call   WaitTimer32
       ld     a,20h
       call   SendControlCmd
       ret    c
       ex     de,hl
       call   GetDescriptor
       ret    c
       ld     a,42h
       jp     SendControlCmd
ReqStringDes:
;A = string index
;HL -> string buffer
       push   hl
       ld     hl,SDReqData
       ld     de,OutBuf
       push   de
       ld     bc,8
       ldir
       ld     (OutBuf+2),a
       pop    hl
       call   Send8PortA0
       pop    hl
       ld     a,0Ah
       call   SendControlCmd
       ret    c
       ld     a,20h
       call   SendControlCmd
       ret    c
       call   GetDescriptor
       ret    c
       ld     a,42h
       jp     SendControlCmd

GetConfigDescriptor:
;Gets the full configuration descriptor
       in     a,(0A0h)
       ld     (hl),a
       inc    hl
       in     a,(0A0h)
       ld     (hl),a
       inc    hl
       in     a,(0A0h)
       ld     (hl),a
       inc    hl
       sub    3
       ld     c,a
       ld     e,0FDh
       jr     GDLoop
GetReportDescriptor:
;Read the incoming report descriptor of size C into buffer HL
       ld     e,0
       jr     GDLoop
GetDescriptor:
;Read the incoming descriptor into buffer HL
       in     a,(0A0h)
       ld     (hl),a
       inc    hl
       dec    a
       ld     c,a
       ld     e,0FFh
GDLoop:
       push   hl
       push   de
       ld     hl,(MemPtr)
       ld     de,offPacketSizeTable
       add    hl,de
       ld     a,(hl)
       pop    de
       pop    hl
       add    a,e
       ld     e,0
       cp     c
       jr     c,$f
       ld     a,c
$$:
       ld     d,a
       call   GetAPortA0
       ld     a,c
       sub    d
       ld     c,a
       jr     z,GDDone
       ld     a,20h
       call   SendControlCmd
       ret    c
       jr     GDLoop
GDDone:
RetAddr:
       or     a
       ret


DDReqDataSmall:
       db     80h,6,0,1,0,0,8,0       
DDReqData:
       db     80h,6,0,1,0,0,18,0
CDReqData:
       db     80h,6,0,2,0,0,128,0
HubDReqData:
       db     0A0h,6,0,29h,0,0,128,0
SDReqData:
       db     80h,6,0,3,9,4,128,0
SetAddrReqData:
       db     0,5,1,0,0,0,0,0
SetConfigReqData:
       db     0,9,0,0,0,0,0,0
GetStatusReqData:
       db     80h,0,0,0,0,0,2,0


Pop2USBErrorNotEnoughMem:
       pop    hl
       pop    hl
USBErrorNotEnoughMem:
       ld     a,errMSDNotEnoughMem
       jr     USBErrorCont

USBErrorFeatureDisabled:
       ld     a,errFeatureDisabled
       jr     USBErrorCont
USBErrorBadCmd:
       ld     a,errBadCmd
       jr     USBErrorCont
USBErrorTimeout:
       ld     a,errTimeout
       jr     USBErrorCont
USBErrorVersion:
       ld     a,errVersion
       jr     USBErrorCont
USBErrorSetAddress:
       ld     a,errSetAddress
       jr     USBErrorCont
USBErrorGetMaxPacketSize:
       ld     a,errGetMaxPacketSize
       jr     USBErrorCont
USBErrorFindPipe:
       ld     a,errFindPipe
       jr     USBErrorCont
USBErrorDemo:
       ld     a,errDemo
       jr     USBErrorCont
USBErrorAutoSetup
       ld     a,errAutoSetup
       jr     USBErrorCont
USBErrorClass:
       ld     a,errBadClass
       jr     USBErrorCont
USBErrorInit:
       ld     a,errInit
       jr     USBErrorCont
PopErrorReadDescriptor:
       pop    hl
USBErrorReadDescriptor:
       ld     a,errReadDescriptor
       jr     USBErrorCont
USBErrorID:
       ld     a,errBadID
       jr     USBErrorCont
PopErrorFindDescriptor:
       pop    hl
USBErrorFindDescriptor:
       ld     a,errFindDescriptor
       jr     USBErrorCont
USBErrorNoDevice:
       ld     a,errNoDevice
       jr     USBErrorCont
USBErrorEPGetDes:
       ld     a,errEPGetDes
       jr     USBErrorCont
USBErrorEPIndex:
       ld     a,errEPIndex
       jr     USBErrorCont
USBErrorPort81:
       ld     a,errPort81
       jr     USBErrorCont
USBErrorPort8C:
       ld     a,errPort8C
       jr     USBErrorCont
USBErrorPort8F:
       ld     a,errPort8F
       jr     USBErrorCont
USBErrorPort3A:
       ld     a,errPort3A
       jr     USBErrorCont
USBErrorPort4D:
       ld     a,errPort4D
       jr     USBErrorCont
USBErrorPort4C:
       ld     a,errPort4C
       jr     USBErrorCont
USBErrorPort91Bit4Pop1:
       pop    hl
USBErrorPort91Bit4:
       ld     a,errPort91Bit4
       jr     USBErrorCont
USBErrorPort91Bit2Pop1:
       pop    hl
USBErrorPort91Bit2:
       ld     a,errPort91Bit2
       jr     USBErrorCont
USBErrorPort82Pop3:
       pop    hl
USBErrorPort82Pop2:
       pop    hl
       pop    hl
USBErrorPort82:
       ld     a,errPort82
       jr     USBErrorCont
;USBError:
;       xor    a      ;errOther
USBErrorCont:
       LOG    Error,a
       ld     hl,(MemPtr)
       ld     de,offErrorCode
       add    hl,de
       ld     b,(hl)
       ld     (hl),a
       inc    hl
       ld     (hl),b
       scf
       ret
