 include "settings.inc"
  include "../ti83plus.inc"
 include "equates.inc" 
 SEGMENT Main
 GLOBALS ON
 EXTERN KillUSBDevice,AutoSetup,USBErrorAutoSetup,USBReqData,DispHexA,USBErrorID,SendControlData
 EXTERN IPutC,GetIDs,GetErrorCode,WaitTimer1,SendControlOutSafe,KBDSetReport,WaitTimer8,WaitTimerB
 EXTERN IGetKey,SetupLog,StopLog,USBErrorTimeout,USBErrorBadCmd,USBErrorFeatureDisabled,DoAssert
 EXTERN ILdHLInd



VernierVersion:
;Output
; DE = size of buffer required for vernier driver
; H = major version
; L = minor version
       ld     hl,VVersion
       ld     de,vernierMemSize
       ret


VernierInit:

 IF VERNIER_ON = 0
       jp     USBErrorFeatureDisabled      
 ELSE

       call   EasyTempInitDo
       ret    nc
       call   KillUSBDevice
       scf
       ret
 ENDIF

 IF VERNIER_ON = 1
EasyTempInitDo:
;Initialize EasyTemp device and driver
;Input
; B = flags
;  bit 0 - set = passthrough callback
; DE -> descriptor buf
; HL -> memory buffer
; ix -> callback routine
;Output
; NC = success
       push   de

       push   hl
       ld     hl,(MemPtr)
       ld     de,offVernierBuf
       add    hl,de
       pop    de
       ld     (hl),e
       inc    hl
       ld     (hl),d                      ;Set vernier buffer pointer in main driver buf


       call   LdDEVBuf
       ld     hl,offvernierCallBack
       add    hl,de
       push   ix
       pop    de
       ld     (hl),e
       inc    hl
       ld     (hl),d                      ;Set callback addr
       inc    hl
       ld     (hl),b                      ;save flags

       

       pop    hl            ;HL->desc buf
       ld     de,EasyTempCB
       push   hl
       call   AutoSetup
       pop    hl
       jp     c,USBErrorAutoSetup

       call   GetIDs
       ret    c
       ld     bc,08F7h
       or     a
       sbc    hl,bc
       jp     nz,USBErrorID
       ld     hl,0002h
       or     a
       sbc    hl,de
       jp     nz,USBErrorID




       ld     b,1
       call   USBReqData


       ld     hl,StartupCommands
       ld     e,4
EasyTempInitLoop:
       ld     b,(hl)
       inc    hl
       ld     c,(hl)
       inc    hl
       ld     d,(hl)
       inc    hl
       push   de
       push   hl
       call   VernierQuickCmd
       pop    hl
       pop    de
       ret    c
       dec    e
       jr     nz,EasyTempInitLoop


       xor    a
       ret

StartupCommands:
       db     vernierCMD_ID_INIT,0,0
       db     vernierCMD_ID_SET_LED_STATE,vernierLED_COLOR_BLACK,vernierLED_BRIGHTNESS_MIN+4
       db     vernierCMD_ID_SET_LED_STATE,vernierLED_COLOR_GREEN,vernierLED_BRIGHTNESS_MIN+4
       db     vernierCMD_ID_START_MEASUREMENTS,0,0

EasyTempCB:
       ld     a,cbNewData
       cp     b      
       jr     nz,CBNotData
       ld     a,d
       or     e
       jr     nz,$f
       LOG    Custom,0FEh
$$:
       call   LdDEVBuf
       ld     hl,(MemPtr)
       ld     a,(hl)
       and    vernierMASK_INPUT_PACKET_TYPE
       cp     vernierINPUT_PACKET_TYPE_MEAS
       jr     nz,CBCmd
       inc    hl
       inc    hl
       ld     c,(hl)
       inc    hl
       ld     b,(hl)                             ;BC = first measurement value in packet
       ld     hl,offvernierMeas
       add    hl,de
       ld     (hl),c
       inc    hl
       ld     (hl),b                             ;Save measurement to buffer

       
CBDoCallback:
       call   LdDEVBuf
       ld     hl,offvernierCallBack
       add    hl,de
       ld     e,(hl)
       inc    hl
       ld     d,(hl)
       inc    hl                                 ;hl-> flags
       bit    vernierFlagCallBack,(hl)
       jr     z,$f
       push   de
       ASSERTC d,9Dh
       ASSERTC 7Fh,d
$$:       

CBDone:
       ld     b,1
       jp     USBReqData

CBCmd:
       ld     bc,8
       ldir                                      ;Save command response to buffer
       jr     CBDoCallback

CBNotData:
       LOG    Custom,0FDh
       jr     CBDone
 ENDIF


VernierToDegree:
;Input
; B = 0 -> to celsius
;     1 -> to fahrenheit
; HL = voltage to convert
;Output
; OP1 = converted value
 IF VERNIER_ON = 0
       ret
 ELSE
       ld     c,0
       bit    7,h
       jr     z,$f
       ld     a,h
       cpl
       ld     h,a
       ld     a,l
       neg
       ld     l,a
       inc    c
$$:
       push   bc
       B_CALL SetXXXXOP2
       pop    bc
       push   bc
       bit    0,c
       jr     z,$f
       B_CALL InvOP2S
$$:
       ld     hl,ConvVal
       rst    20h
       B_CALL FPMult
       pop    bc
       bit    0,b
       ret    z                           ;Done if conv to celsius
       ld     a,9
       B_CALL SetXXOP2
       B_CALL FPMult
       B_CALL OP2Set5
       B_CALL FPDiv
       ld     a,32
       B_CALL SetXXOP2
       B_CALL FPAdd
       ret
ConvVal:
       db     00,07Dh,78h,12h,50h,00,00,00,00                  ;.0078125
 ENDIF

VernierGetMeas:
;Output
; HL = current measurement
 IF VERNIER_ON = 0
       ret
 ELSE
       call   LdDEVBuf
       ld     hl,offvernierMeas
       add    hl,de
       jp     ILdHLInd
 ENDIF

VernierQuickCmd:
;Easy send comamnd to vernier
;Input
; B = Command
; C = Param1
; D = Param2
;Output
; DE -> command response
; NC = success
 IF VERNIER_ON = 0
       jp     USBErrorFeatureDisabled      
 ELSE
       push   bc
       push   de
       B_CALL ZeroOP1
       ld     hl,OP1
       pop    de
       pop    bc
       ld     (hl),b
       inc    hl
       ld     (hl),c
       inc    hl
       ld     (hl),d
       ld     de,OP1
 ENDIF

VernierSendCmd:
;Send command to vernier
;Input
; DE -> command block
;Output
; DE -> command response
; NC = success
 IF VERNIER_ON = 0
       jp     USBErrorFeatureDisabled      
 ELSE
       push   bc
       push   hl
       push   de

       push   de
       call   LdDEVBuf
       inc    de
       xor    a
       ld     (de),a                      ;clear response command so we don't read old command below
       pop    de




       
       ld     a,(de)                      ;Save cmd to check for in response
       ld     b,a

       sub    vernierCMD_ID_WRITE_LOCAL_NV_MEM_1BYTE
       cp     6
       ld     a,b
       jr     c,SendCommandBad            ;Don't allow write commands

       push   af
       ld     b,8
       ld     hl,VernierSetReport
       call   SendControlOutSafe
       pop    bc
       jr     c,VernierSendCmdDone
       
       call   LdDEVBuf

       inc    de
       ld     hl,0FFFFh                   ;a while
VernierSendCmdLoop:
       ld     a,vernierCMD_ID_INIT
       cp     b                           ;Did we send an init command
       jr     nz,$f                       ; if not continue as normal
       dec    de
       ld     a,(de)
       inc    de
       and    vernierINPUT_PACKET_INIT_RESP
       jr     nz,VernierSendCmdFound
$$:
       ld     a,(de)
       cp     b                           ;If received cmd = sent cmd
       jr     z,VernierSendCmdFound       ; done w/success 
       dec    hl
       ld     a,h
       or     l
       jr     nz,VernierSendCmdLoop
       call   USBErrorTimeout             ;Didn't get cmd back from device

VernierSendCmdFound:
       dec    de                          ;de -> header
       ld     a,(de)
       and    vernierMASK_INPUT_PACKET_ERROR_FLAG          ;Error response
       jr     z,$f                        ;if no error, continue w/o carry
SendCommandBad:
       call   USBErrorBadCmd              ;set error/carry
$$:

VernierSendCmdDone:
       pop    de
       pop    hl
       pop    bc
       push   af
       call   LdDEVBuf
       pop    af
       ret



LdDEVBuf:
       push   hl
       ld     hl,(MemPtr)
       ASSERTC h,9Dh
       ASSERTC 7Fh,h
       ld     de,offVernierBuf
       add    hl,de
       ld     e,(hl)
       inc    hl
       ld     d,(hl)
       ASSERTC d,9Dh
       ASSERTC 7Fh,d
       pop    hl
       ret



VernierSetReport:
       db     21h,9,00,02,00,00,8,00

 ENDIF
